# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Greedy sampler."""

import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_nlp.samplers.greedy_sampler import GreedySampler


class GreedySamplerTest(tf.test.TestCase, parameterized.TestCase):
    def setUp(self):
        super().setUp()
        self.vocab_size = 10
        self.feature_size = 16

        # Create a dummy model to predict the next token.
        model = keras.Sequential(
            [
                keras.Input(shape=[None]),
                keras.layers.Embedding(
                    input_dim=self.vocab_size,
                    output_dim=self.feature_size,
                ),
                keras.layers.Dense(self.vocab_size),
                keras.layers.Softmax(),
            ]
        )

        def token_probability_fn(inputs, mask):
            return model(inputs)

        self.token_probability_fn = token_probability_fn

        self.sampler = GreedySampler()

    def test_generate_with_1d_prompt(self):
        inputs = tf.constant([1])
        outputs = self.sampler(inputs, self.token_probability_fn, max_length=5)
        self.assertEqual(outputs.shape, [5])

    def test_generate_with_2d_prompt(self):
        inputs = tf.constant([[1], [1]])
        outputs = self.sampler(inputs, self.token_probability_fn, max_length=5)
        self.assertEqual(outputs.shape, [2, 5])

    def test_generate_with_list_prompt(self):
        inputs = [[1], [1]]
        outputs = self.sampler(inputs, self.token_probability_fn, max_length=5)
        self.assertEqual(outputs.shape, [2, 5])

    def test_generate_with_ragged_prompt(self):
        max_length = 5

        def token_probability_fn(inputs, mask):
            # Assert that user function is passed only dense tensors.
            self.assertIsInstance(inputs, tf.Tensor)
            prob = tf.constant([[[0.0, 0.0, 0.0, 1.0]]])
            return tf.repeat(tf.repeat(prob, 2, axis=0), max_length, axis=1)

        inputs = tf.ragged.constant([[1], [2, 1, 2]])
        outputs = self.sampler(inputs, token_probability_fn, max_length)
        self.assertEqual(outputs.shape, [2, 5])

    def test_assert_generation_is_correct(self):
        batch_size = 10
        max_length = 3

        def token_probability_fn(inputs, mask):
            prob = tf.constant([[[0.0, 0.0, 0.0, 1.0]]])
            return tf.repeat(
                tf.repeat(prob, batch_size, axis=0), max_length, axis=1
            )

        inputs = 3 * tf.ones([batch_size, 1], dtype=tf.int32)
        outputs = self.sampler(
            inputs, token_probability_fn, max_length=max_length
        )
        self.assertAllEqual(
            outputs, 3 * tf.ones(shape=[batch_size, max_length])
        )

    def test_end_token_id(self):
        def token_probability_fn(inputs, mask):
            batch_size = inputs.shape[0]
            prob = tf.constant([[[0.0, 0.0, 0.0, 1.0]]])
            return tf.repeat(
                tf.repeat(prob, batch_size, axis=0), max_length, axis=1
            )

        max_length = 4
        sampler = GreedySampler()
        inputs = tf.constant([[0, 1], [1, 2]])
        outputs = sampler(
            inputs,
            token_probability_fn,
            max_length=max_length,
            end_token_id=2,
        )
        # end_token in prompt does not trigger truncation.
        expected_outputs = tf.ragged.constant([[0, 1, 3, 3], [1, 2, 3, 3]])
        self.assertAllEqual(outputs, expected_outputs)

        outputs = sampler(
            inputs,
            token_probability_fn,
            max_length=max_length,
            end_token_id=3,
        )
        # Generated end_token will be truncated.
        expected_outputs = tf.ragged.constant([[0, 1], [1, 2]])
        self.assertAllEqual(outputs, expected_outputs)

    def test_compare_xla_noxla_results(self):
        inputs = [[1], [1]]
        xla_sampler = GreedySampler(jit_compile=True)
        outputs_xla = xla_sampler(
            inputs, self.token_probability_fn, max_length=5
        )

        xla_sampler = GreedySampler(jit_compile=False)
        outputs_no_xla = xla_sampler(
            inputs, self.token_probability_fn, max_length=5
        )

        self.assertAllEqual(outputs_xla, outputs_no_xla)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for SentencePiece Tokenizer Trainer."""

import os
import re

import tensorflow as tf

from keras_nlp.tokenizers.sentence_piece_tokenizer import SentencePieceTokenizer
from keras_nlp.tokenizers.sentence_piece_tokenizer_trainer import (
    compute_sentence_piece_proto,
)


class SentencePieceTokenizerTrainerTest(tf.test.TestCase):
    def test_dataset_input(self):
        test_text = ["Ninjas and Samurais"]
        expected_output = [
            [5, 9, 6, 7, 11, 4, 8, 5, 4, 7, 10, 5, 12, 4, 13, 15, 14, 4, 6, 8]
        ]
        data = tf.data.Dataset.from_tensor_slices(test_text)
        proto = compute_sentence_piece_proto(data, 16)
        tokenizer = SentencePieceTokenizer(proto=proto)
        test_output = tokenizer(test_text)
        self.assertAllEqual(expected_output, test_output)

    def test_file_input(self):
        test_text = "Ninja Land"
        with open(os.path.join(self.get_temp_dir(), "test.txt"), "w+") as f:
            f.write(test_text + "\n")
        expected_output = [6, 8, 9, 5, 11, 4, 6, 7, 4, 5, 10]
        proto = compute_sentence_piece_proto(
            [os.path.join(self.get_temp_dir(), "test.txt")],
            12,
        )
        tokenizer = SentencePieceTokenizer(proto=proto)
        test_output = tokenizer(test_text)
        self.assertAllEqual(expected_output, test_output)

    def test_multiple_file_input(self):
        with open(os.path.join(self.get_temp_dir(), "test1.txt"), "w+") as f:
            f.write("Drifting Along\n")
        with open(os.path.join(self.get_temp_dir(), "test2.txt"), "w+") as f:
            f.write("Woah look there\n")
        inputs = [
            os.path.join(self.get_temp_dir(), "test1.txt"),
            os.path.join(self.get_temp_dir(), "test2.txt"),
        ]
        proto = compute_sentence_piece_proto(inputs, 20)
        tokenizer = SentencePieceTokenizer(proto=proto)
        test_text = "Woah Along"
        test_output = tokenizer(test_text)
        expected_output = [4, 16, 5, 17, 9, 4, 15, 12, 5, 11, 18]
        self.assertAllEqual(expected_output, test_output)

    def test_invalid_input(self):
        test_text_invalid = {"file1": "test.txt"}
        with self.assertRaisesRegex(
            ValueError,
            re.escape(f"Received: type(data)={type(test_text_invalid)}."),
        ):
            compute_sentence_piece_proto(test_text_invalid, 10)

    def test_lowercase(self):
        inputs = tf.data.Dataset.from_tensor_slices(["Drifting Along"])
        proto = compute_sentence_piece_proto(
            inputs, vocabulary_size=15, lowercase=True
        )
        tokenizer = SentencePieceTokenizer(proto=proto)
        output = inputs.map(tokenizer).take(1).get_single_element()
        expected_output = [4, 8, 12, 5, 9, 14, 5, 6, 13, 4, 7, 10, 11, 6, 13]
        self.assertAllEqual(expected_output, output)

    def test_proto_output_file(self):
        inputs = tf.data.Dataset.from_tensor_slices(["Drifting Along"])
        compute_sentence_piece_proto(
            inputs,
            vocabulary_size=15,
            proto_output_file=os.path.join(self.get_temp_dir(), "model.spm"),
        )
        tokenizer = SentencePieceTokenizer(
            proto=os.path.join(self.get_temp_dir(), "model.spm")
        )
        output = inputs.map(tokenizer).take(1).get_single_element()
        expected_output = [4, 8, 12, 5, 9, 14, 5, 6, 13, 4, 7, 10, 11, 6, 13]
        self.assertAllEqual(expected_output, output)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for DeBERTa tokenizer."""

import io
import os

import sentencepiece
import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_nlp.models.deberta_v3.deberta_v3_tokenizer import DebertaV3Tokenizer


class DebertaV3TokenizerTest(tf.test.TestCase, parameterized.TestCase):
    def setUp(self):
        bytes_io = io.BytesIO()
        vocab_data = tf.data.Dataset.from_tensor_slices(
            ["the quick brown fox", "the earth is round"]
        )
        sentencepiece.SentencePieceTrainer.train(
            sentence_iterator=vocab_data.as_numpy_iterator(),
            model_writer=bytes_io,
            vocab_size=10,
            model_type="WORD",
            pad_id=0,
            bos_id=1,
            eos_id=2,
            unk_id=3,
            pad_piece="[PAD]",
            bos_piece="[CLS]",
            eos_piece="[SEP]",
            unk_piece="[UNK]",
        )
        self.proto = bytes_io.getvalue()

        self.tokenizer = DebertaV3Tokenizer(proto=self.proto)

    def test_tokenize(self):
        input_data = "the quick brown fox"
        output = self.tokenizer(input_data)
        self.assertAllEqual(output, [4, 9, 5, 7])

    def test_tokenize_batch(self):
        input_data = tf.constant(["the quick brown fox", "the earth is round"])
        output = self.tokenizer(input_data)
        self.assertAllEqual(output, [[4, 9, 5, 7], [4, 6, 8, 3]])

    def test_detokenize(self):
        input_data = tf.constant([[4, 9, 5, 7]])
        output = self.tokenizer.detokenize(input_data)
        self.assertEqual(output, tf.constant(["the quick brown fox"]))

    def test_vocabulary_size(self):
        self.assertEqual(self.tokenizer.vocabulary_size(), 10)

    def test_errors_missing_special_tokens(self):
        bytes_io = io.BytesIO()
        sentencepiece.SentencePieceTrainer.train(
            sentence_iterator=iter(["abc"]),
            model_writer=bytes_io,
            vocab_size=5,
            pad_id=-1,
            eos_id=-1,
            bos_id=-1,
        )
        with self.assertRaises(ValueError):
            DebertaV3Tokenizer(proto=bytes_io.getvalue())

    @parameterized.named_parameters(
        ("tf_format", "tf", "model"),
        ("keras_format", "keras_v3", "model.keras"),
    )
    def test_saved_model(self, save_format, filename):
        input_data = tf.constant(["the quick brown fox"])

        inputs = keras.Input(dtype="string", shape=())
        outputs = self.tokenizer(inputs)
        model = keras.Model(inputs, outputs)

        path = os.path.join(self.get_temp_dir(), filename)
        model.save(path, save_format=save_format)

        restored_model = keras.models.load_model(path)
        self.assertAllEqual(
            model(input_data),
            restored_model(input_data),
        )



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import shutil

import numpy as np
import tensorflow as tf
import torch
import transformers
from absl import app
from absl import flags
from tensorflow import keras

import keras_nlp
from tools.checkpoint_conversion.checkpoint_conversion_utils import (
    get_md5_checksum,
)

PRESET_MAP = {
    "roberta_base": ("roberta.base", "roberta-base"),
    "roberta_large": ("roberta.large", "roberta-large"),
}

DOWNLOAD_SCRIPT_URL = "https://dl.fbaipublicfiles.com/fairseq/models/{}.tar.gz"

EXTRACT_DIR = "./{}"

FLAGS = flags.FLAGS
flags.DEFINE_string(
    "preset", None, f'Must be one of {",".join(PRESET_MAP.keys())}'
)


def download_model(preset, size, hf_model_name):
    print("-> Download original weights.")
    extract_dir = EXTRACT_DIR.format(size)
    archive_file_path = keras.utils.get_file(
        fname=None,
        origin=DOWNLOAD_SCRIPT_URL.format(size),
        cache_subdir=os.path.join("checkpoint_conversion", preset),
    )

    os.system(f"tar -xvf {archive_file_path}")

    # The original `tar.gz` file does not have the vocab files. Let's fetch
    # them from HF.
    vocabulary_path = keras.utils.get_file(
        fname=None,
        origin=f"https://huggingface.co/{hf_model_name}/raw/main/vocab.json",
    )
    shutil.copy(vocabulary_path, extract_dir)
    merges_path = keras.utils.get_file(
        fname=None,
        origin=f"https://huggingface.co/{hf_model_name}/raw/main/merges.txt",
    )
    shutil.copy(merges_path, extract_dir)


def convert_checkpoints(preset, size):
    print("\n-> Convert original weights to KerasNLP format.")
    # RoBERTa paths.
    extract_dir = EXTRACT_DIR.format(size)
    checkpoint_path = os.path.join(extract_dir, "model.pt")

    # Load PyTorch RoBERTa checkpoint.
    pt_ckpt = torch.load(checkpoint_path, map_location=torch.device("cpu"))
    pt_cfg = pt_ckpt["args"]
    pt_model = pt_ckpt["model"]

    cfg = {
        "num_layers": pt_cfg.encoder_layers,
        "num_heads": pt_cfg.encoder_attention_heads,
        "hidden_dim": pt_cfg.encoder_embed_dim,
        "intermediate_dim": pt_cfg.encoder_ffn_embed_dim,
        "dropout": pt_cfg.dropout,
        "max_sequence_length": pt_cfg.max_positions,
        "vocab_size": (
            pt_model["decoder.sentence_encoder.embed_tokens.weight"]
            .numpy()
            .shape[0]
        ),
    }
    print("Config:", cfg)

    keras_nlp_model = keras_nlp.models.RobertaBackbone.from_preset(
        preset, load_weights=False
    )

    # Embedding Layer.
    keras_nlp_model.get_layer("embeddings").token_embedding.embeddings.assign(
        pt_model["decoder.sentence_encoder.embed_tokens.weight"].numpy()
    )
    keras_nlp_model.get_layer(
        "embeddings"
    ).position_embedding.position_embeddings.assign(
        pt_model["decoder.sentence_encoder.embed_positions.weight"].numpy()[
            2:, :
        ]
    )

    # Embedding LayerNorm.
    keras_nlp_model.get_layer("embeddings_layer_norm").gamma.assign(
        pt_model["decoder.sentence_encoder.emb_layer_norm.weight"].numpy()
    )
    keras_nlp_model.get_layer("embeddings_layer_norm").beta.assign(
        pt_model["decoder.sentence_encoder.emb_layer_norm.bias"].numpy()
    )

    # The QKV weights in the original checkpoint are present as one single
    # dense layer of shape `(3 * cfg["hidden_dim"], cfg["hidden_dim"])`. Our
    # model has three separate dense layers for each of QKV. Hence, we need to
    # split the original QKV weights into three chunks.
    range_1 = (0, cfg["hidden_dim"])
    range_2 = (cfg["hidden_dim"], 2 * cfg["hidden_dim"])
    range_3 = (2 * cfg["hidden_dim"], 3 * cfg["hidden_dim"])
    # Transformer layers.
    for i in range(keras_nlp_model.num_layers):
        q_k_v_wts = (
            pt_model[
                f"decoder.sentence_encoder.layers.{i}.self_attn.in_proj_weight"
            ]
            .numpy()
            .T
        )
        q_k_v_bias = (
            pt_model[
                f"decoder.sentence_encoder.layers.{i}.self_attn.in_proj_bias"
            ]
            .numpy()
            .T
        )

        # Query
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._query_dense.kernel.assign(
            q_k_v_wts[:, range_1[0] : range_1[1]].reshape(
                (cfg["hidden_dim"], cfg["num_heads"], -1)
            )
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._query_dense.bias.assign(
            q_k_v_bias[range_1[0] : range_1[1]].reshape((cfg["num_heads"], -1))
        )

        # Key
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._key_dense.kernel.assign(
            q_k_v_wts[:, range_2[0] : range_2[1]].reshape(
                (cfg["hidden_dim"], cfg["num_heads"], -1)
            )
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._key_dense.bias.assign(
            q_k_v_bias[range_2[0] : range_2[1]].reshape((cfg["num_heads"], -1))
        )

        # Value
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._value_dense.kernel.assign(
            q_k_v_wts[:, range_3[0] : range_3[1]].reshape(
                (cfg["hidden_dim"], cfg["num_heads"], -1)
            )
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._value_dense.bias.assign(
            q_k_v_bias[range_3[0] : range_3[1]].reshape((cfg["num_heads"], -1))
        )

        # Attention output
        attn_output_wts = (
            pt_model[
                f"decoder.sentence_encoder.layers.{i}.self_attn.out_proj.weight"
            ]
            .numpy()
            .T
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._output_dense.kernel.assign(
            attn_output_wts.reshape((cfg["num_heads"], -1, cfg["hidden_dim"]))
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._output_dense.bias.assign(
            pt_model[
                f"decoder.sentence_encoder.layers.{i}.self_attn.out_proj.bias"
            ].numpy()
        )

        # Attention LayerNorm
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layernorm.gamma.assign(
            pt_model[
                f"decoder.sentence_encoder.layers.{i}.self_attn_layer_norm.weight"
            ].numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layernorm.beta.assign(
            pt_model[
                f"decoder.sentence_encoder.layers.{i}.self_attn_layer_norm.bias"
            ].numpy()
        )

        # Intermediate FF layer
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_intermediate_dense.kernel.assign(
            pt_model[f"decoder.sentence_encoder.layers.{i}.fc1.weight"]
            .numpy()
            .T
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_intermediate_dense.bias.assign(
            pt_model[f"decoder.sentence_encoder.layers.{i}.fc1.bias"].numpy()
        )

        # Output dense layer
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_output_dense.kernel.assign(
            pt_model[f"decoder.sentence_encoder.layers.{i}.fc2.weight"]
            .numpy()
            .T
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_output_dense.bias.assign(
            pt_model[f"decoder.sentence_encoder.layers.{i}.fc2.bias"].numpy()
        )

        # FF LayerNorm
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_layernorm.gamma.assign(
            pt_model[
                f"decoder.sentence_encoder.layers.{i}.final_layer_norm.weight"
            ].numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_layernorm.beta.assign(
            pt_model[
                f"decoder.sentence_encoder.layers.{i}.final_layer_norm.bias"
            ].numpy()
        )

    # Save the model.
    print(f"\n-> Save KerasNLP model weights to `{preset}.h5`.")
    keras_nlp_model.save_weights(f"{preset}.h5")

    return keras_nlp_model


def define_preprocessor(preset, hf_model_name, size):
    print("\n-> Define the tokenizers.")
    extract_dir = EXTRACT_DIR.format(size)
    vocabulary_path = os.path.join(extract_dir, "vocab.json")
    merges_path = os.path.join(extract_dir, "merges.txt")

    keras_nlp_tokenizer = keras_nlp.models.RobertaTokenizer(
        vocabulary=vocabulary_path, merges=merges_path
    )
    keras_nlp_preprocessor = keras_nlp.models.RobertaPreprocessor(
        keras_nlp_tokenizer
    )

    hf_tokenizer = transformers.AutoTokenizer.from_pretrained(hf_model_name)

    print("\n-> Print MD5 checksum of the vocab files.")
    print(f"`{vocabulary_path}` md5sum: ", get_md5_checksum(vocabulary_path))
    print(f"`{merges_path}` md5sum: ", get_md5_checksum(merges_path))

    return keras_nlp_preprocessor, hf_tokenizer


def check_output(
    preset,
    keras_nlp_model,
    keras_nlp_preprocessor,
    hf_model,
    hf_tokenizer,
):
    print("\n-> Check the outputs.")
    input_str = ["the quick brown fox ran, galloped and jumped."]

    # KerasNLP
    keras_nlp_inputs = keras_nlp_preprocessor(tf.constant(input_str))
    keras_nlp_output = keras_nlp_model.predict(keras_nlp_inputs)

    # HF
    hf_inputs = hf_tokenizer(
        input_str, padding="max_length", return_tensors="pt"
    )
    hf_output = hf_model(**hf_inputs).last_hidden_state

    print("KerasNLP output:", keras_nlp_output[0, 0, :10])
    print("HF output:", hf_output[0, 0, :10])
    print("Difference:", np.mean(keras_nlp_output - hf_output.detach().numpy()))

    # Show the MD5 checksum of the model weights.
    print("Model md5sum: ", get_md5_checksum(f"./{preset}.h5"))

    return keras_nlp_output


def main(_):
    assert (
        FLAGS.preset in PRESET_MAP.keys()
    ), f'Invalid preset {FLAGS.preset}. Must be one of {",".join(PRESET_MAP.keys())}'
    size = PRESET_MAP[FLAGS.preset][0]
    hf_model_name = PRESET_MAP[FLAGS.preset][1]

    download_model(FLAGS.preset, size, hf_model_name)

    keras_nlp_model = convert_checkpoints(FLAGS.preset, size)

    print("\n-> Load HF model.")
    hf_model = transformers.AutoModel.from_pretrained(hf_model_name)
    hf_model.eval()

    keras_nlp_preprocessor, hf_tokenizer = define_preprocessor(
        FLAGS.preset, hf_model_name, size
    )

    check_output(
        FLAGS.preset,
        keras_nlp_model,
        keras_nlp_preprocessor,
        hf_model,
        hf_tokenizer,
    )


if __name__ == "__main__":
    flags.mark_flag_as_required("preset")
    app.run(main)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for BERT tokenizer."""

import os

import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_nlp.models.bert.bert_tokenizer import BertTokenizer


class BertTokenizerTest(tf.test.TestCase, parameterized.TestCase):
    def setUp(self):
        self.vocab = ["[PAD]", "[UNK]", "[CLS]", "[SEP]", "[MASK]"]
        self.vocab += ["THE", "QUICK", "BROWN", "FOX"]
        self.vocab += ["the", "quick", "brown", "fox"]
        self.tokenizer = BertTokenizer(vocabulary=self.vocab)

    def test_tokenize(self):
        input_data = "THE QUICK BROWN FOX."
        output = self.tokenizer(input_data)
        self.assertAllEqual(output, [5, 6, 7, 8, 1])

    def test_tokenize_batch(self):
        input_data = tf.constant(["THE QUICK BROWN FOX.", "THE FOX."])
        output = self.tokenizer(input_data)
        self.assertAllEqual(output, [[5, 6, 7, 8, 1], [5, 8, 1]])

    def test_lowercase(self):
        input_data = "THE QUICK BROWN FOX."
        tokenizer = BertTokenizer(vocabulary=self.vocab, lowercase=True)
        output = tokenizer(input_data)
        self.assertAllEqual(output, [9, 10, 11, 12, 1])

    def test_detokenize(self):
        input_tokens = [[5, 6, 7, 8]]
        output = self.tokenizer.detokenize(input_tokens)
        self.assertAllEqual(output, ["THE QUICK BROWN FOX"])

    def test_vocabulary_size(self):
        self.assertEqual(self.tokenizer.vocabulary_size(), 13)

    def test_errors_missing_special_tokens(self):
        with self.assertRaises(ValueError):
            BertTokenizer(vocabulary=["a", "b", "c"])

    @parameterized.named_parameters(
        ("tf_format", "tf", "model"),
        ("keras_format", "keras_v3", "model.keras"),
    )
    def test_saved_model(self, save_format, filename):
        input_data = tf.constant(["THE QUICK BROWN FOX."])
        tokenizer = BertTokenizer(vocabulary=self.vocab)
        inputs = keras.Input(dtype="string", shape=())
        outputs = tokenizer(inputs)
        model = keras.Model(inputs, outputs)
        path = os.path.join(self.get_temp_dir(), filename)
        model.save(path, save_format=save_format)
        restored_model = keras.models.load_model(path)
        self.assertAllEqual(
            model(input_data),
            restored_model(input_data),
        )



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import time

import tensorflow as tf
import tensorflow_datasets as tfds
from absl import app
from absl import flags
from tensorflow import keras

import keras_nlp

FLAGS = flags.FLAGS
flags.DEFINE_string(
    "model",
    None,
    "The name of the classifier such as BertClassifier.",
)
flags.DEFINE_string(
    "preset",
    None,
    "The name of a preset, e.g. bert_base_multi.",
)

flags.DEFINE_string(
    "mixed_precision_policy",
    "mixed_float16",
    "The global precision policy to use. E.g. 'mixed_float16' or 'float32'.",
)

flags.DEFINE_float("learning_rate", 5e-5, "The learning rate.")
flags.DEFINE_integer("num_epochs", 1, "The number of epochs.")
flags.DEFINE_integer("batch_size", 16, "The batch size.")

tfds.disable_progress_bar()

BUFFER_SIZE = 10000


def create_imdb_dataset():
    dataset, info = tfds.load(
        "imdb_reviews", as_supervised=True, with_info=True
    )
    train_dataset, test_dataset = dataset["train"], dataset["test"]

    train_dataset = (
        train_dataset.shuffle(BUFFER_SIZE)
        .batch(FLAGS.batch_size)
        .prefetch(tf.data.AUTOTUNE)
    )

    # We split the test data evenly into validation and test sets.
    test_dataset_size = info.splits["test"].num_examples // 2

    val_dataset = (
        test_dataset.take(test_dataset_size)
        .batch(FLAGS.batch_size)
        .prefetch(tf.data.AUTOTUNE)
    )
    test_dataset = (
        test_dataset.skip(test_dataset_size)
        .batch(FLAGS.batch_size)
        .prefetch(tf.data.AUTOTUNE)
    )

    return train_dataset, val_dataset, test_dataset


def create_model():
    for name, symbol in keras_nlp.models.__dict__.items():
        if inspect.isclass(symbol) and issubclass(symbol, keras.Model):
            if FLAGS.model and name != FLAGS.model:
                continue
            if not hasattr(symbol, "from_preset"):
                continue
            for preset in symbol.presets:
                if FLAGS.preset and preset != FLAGS.preset:
                    continue
                model = symbol.from_preset(preset)
                print(f"Using model {name} with preset {preset}")
                return model

    raise ValueError(f"Model {FLAGS.model} or preset {FLAGS.preset} not found.")


def train_model(
    model: keras.Model,
    train_dataset: tf.data.Dataset,
    validation_dataset: tf.data.Dataset,
):
    model.compile(
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        optimizer=keras.optimizers.Adam(5e-5),
        metrics=keras.metrics.SparseCategoricalAccuracy(),
        jit_compile=True,
    )

    model.fit(
        train_dataset,
        epochs=FLAGS.num_epochs,
        validation_data=validation_dataset,
        verbose=2,
    )

    return model


def evaluate_model(model: keras.Model, test_dataset: tf.data.Dataset):
    loss, accuracy = model.evaluate(test_dataset)
    print(f"Test loss: {loss}")
    print(f"Test accuracy: {accuracy}")


def main(_):
    keras.mixed_precision.set_global_policy(FLAGS.mixed_precision_policy)

    # Start time
    start_time = time.time()

    train_dataset, validation_dataset, test_dataset = create_imdb_dataset()
    model = create_model()
    model = train_model(model, train_dataset, validation_dataset)
    evaluate_model(model, test_dataset)

    # End time
    end_time = time.time()
    print(f"Total wall time: {end_time - start_time}")


if __name__ == "__main__":
    flags.mark_flag_as_required("model")
    app.run(main)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Greedy Sampler."""

import tensorflow as tf
from tensorflow import keras

from keras_nlp.samplers.sampler import Sampler
from keras_nlp.samplers.sampler import base_sampler_args_docstring
from keras_nlp.samplers.sampler import call_args_docstring
from keras_nlp.utils.python_utils import format_docstring


@format_docstring(
    base_sampler_args=base_sampler_args_docstring, call_args=call_args_docstring
)
@keras.utils.register_keras_serializable(package="keras_nlp")
class GreedySampler(Sampler):
    """Greedy sampler class.

    This sampler is implemented on greedy search, i.e., always picking up the
    token of the largest probability as the next token.

    Args:
        {{base_sampler_args}}

    Call Args:
        {{call_args}}

    Examples:
    ```python
    VOCAB_SIZE = 10

    # Create a dummy model to predict the next token.
    model = keras.Sequential(
        [
            keras.Input(shape=[None]),
            keras.layers.Embedding(
                input_dim=VOCAB_SIZE,
                output_dim=16,
            ),
            keras.layers.Dense(VOCAB_SIZE, activation="softmax"),
        ]
    )

    # Define a function that outputs the next token's probability for each token
    # in the input sequence.
    def token_probability_fn(inputs, mask):
        return model(inputs)

    prompt = tf.fill((8, 1), 1)

    sampler = keras_nlp.samplers.GreedySampler()
    # Print the generated sequence (token ids).
    print(sampler(prompt, token_probability_fn, max_length=10))
    ```
    """

    def __init__(
        self,
        jit_compile=True,
        run_eagerly=False,
    ):
        super().__init__(jit_compile=jit_compile, run_eagerly=run_eagerly)

    def get_next_token(self, next_token_probs):
        return tf.argmax(next_token_probs, axis=-1)



# Copyright 2022 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for RoBERTa masked language model preprocessor layer."""

import os

import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_nlp.models.roberta.roberta_masked_lm_preprocessor import (
    RobertaMaskedLMPreprocessor,
)
from keras_nlp.models.roberta.roberta_tokenizer import RobertaTokenizer


class RobertaMaskedLMPreprocessorTest(tf.test.TestCase, parameterized.TestCase):
    def setUp(self):
        vocab = {
            "<s>": 0,
            "<pad>": 1,
            "</s>": 2,
            "Ġair": 3,
            "plane": 4,
            "Ġat": 5,
            "port": 6,
            "Ġkoh": 7,
            "li": 8,
            "Ġis": 9,
            "Ġthe": 10,
            "Ġbest": 11,
            "<mask>": 12,
        }

        merges = ["Ġ a", "Ġ t", "Ġ k", "Ġ i", "Ġ b", "Ġa i", "p l", "n e"]
        merges += ["Ġa t", "p o", "r t", "o h", "l i", "Ġi s", "Ġb e", "s t"]
        merges += ["Ġt h", "Ġai r", "pl a", "Ġk oh", "Ġth e", "Ġbe st", "po rt"]
        merges += ["pla ne"]

        self.preprocessor = RobertaMaskedLMPreprocessor(
            tokenizer=RobertaTokenizer(
                vocabulary=vocab,
                merges=merges,
            ),
            # Simplify out testing by masking every available token.
            mask_selection_rate=1.0,
            mask_token_rate=1.0,
            random_token_rate=0.0,
            mask_selection_length=5,
            sequence_length=12,
        )

    def test_preprocess_strings(self):
        input_data = " airplane at airport"

        x, y, sw = self.preprocessor(input_data)
        self.assertAllEqual(
            x["token_ids"], [0, 12, 12, 12, 12, 12, 2, 1, 1, 1, 1, 1]
        )
        self.assertAllEqual(
            x["padding_mask"], [1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0]
        )
        self.assertAllEqual(x["mask_positions"], [1, 2, 3, 4, 5])
        self.assertAllEqual(y, [3, 4, 5, 3, 6])
        self.assertAllEqual(sw, [1.0, 1.0, 1.0, 1.0, 1.0])

    def test_preprocess_list_of_strings(self):
        input_data = [" airplane at airport"] * 4

        x, y, sw = self.preprocessor(input_data)
        self.assertAllEqual(
            x["token_ids"], [[0, 12, 12, 12, 12, 12, 2, 1, 1, 1, 1, 1]] * 4
        )
        self.assertAllEqual(
            x["padding_mask"], [[1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0]] * 4
        )
        self.assertAllEqual(x["mask_positions"], [[1, 2, 3, 4, 5]] * 4)
        self.assertAllEqual(y, [[3, 4, 5, 3, 6]] * 4)
        self.assertAllEqual(sw, [[1.0, 1.0, 1.0, 1.0, 1.0]] * 4)

    def test_preprocess_dataset(self):
        sentences = tf.constant([" airplane at airport"] * 4)
        ds = tf.data.Dataset.from_tensor_slices(sentences)
        ds = ds.map(self.preprocessor)
        x, y, sw = ds.batch(4).take(1).get_single_element()
        self.assertAllEqual(
            x["token_ids"], [[0, 12, 12, 12, 12, 12, 2, 1, 1, 1, 1, 1]] * 4
        )
        self.assertAllEqual(
            x["padding_mask"], [[1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0]] * 4
        )
        self.assertAllEqual(x["mask_positions"], [[1, 2, 3, 4, 5]] * 4)
        self.assertAllEqual(y, [[3, 4, 5, 3, 6]] * 4)
        self.assertAllEqual(sw, [[1.0, 1.0, 1.0, 1.0, 1.0]] * 4)

    def test_mask_multiple_sentences(self):
        sentence_one = tf.constant(" airplane")
        sentence_two = tf.constant(" kohli")

        x, y, sw = self.preprocessor((sentence_one, sentence_two))
        self.assertAllEqual(
            x["token_ids"], [0, 12, 12, 2, 2, 12, 12, 2, 1, 1, 1, 1]
        )
        self.assertAllEqual(
            x["padding_mask"], [1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0]
        )
        self.assertAllEqual(x["mask_positions"], [1, 2, 5, 6, 0])
        self.assertAllEqual(y, [3, 4, 7, 8, 0])
        self.assertAllEqual(sw, [1.0, 1.0, 1.0, 1.0, 0.0])

    def test_no_masking_zero_rate(self):
        no_mask_preprocessor = RobertaMaskedLMPreprocessor(
            self.preprocessor.tokenizer,
            mask_selection_rate=0.0,
            mask_selection_length=5,
            sequence_length=12,
        )
        input_data = " airplane at airport"

        x, y, sw = no_mask_preprocessor(input_data)
        self.assertAllEqual(
            x["token_ids"], [0, 3, 4, 5, 3, 6, 2, 1, 1, 1, 1, 1]
        )
        self.assertAllEqual(
            x["padding_mask"], [1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0]
        )
        self.assertAllEqual(x["mask_positions"], [0, 0, 0, 0, 0])
        self.assertAllEqual(y, [0, 0, 0, 0, 0])
        self.assertAllEqual(sw, [0.0, 0.0, 0.0, 0.0, 0.0])

    @parameterized.named_parameters(
        ("tf_format", "tf", "model"),
        ("keras_format", "keras_v3", "model.keras"),
    )
    def test_saved_model(self, save_format, filename):
        input_data = tf.constant([" airplane at airport"])

        inputs = keras.Input(dtype="string", shape=())
        outputs = self.preprocessor(inputs)
        model = keras.Model(inputs, outputs)

        path = os.path.join(self.get_temp_dir(), filename)
        model.save(path, save_format=save_format)

        restored_model = keras.models.load_model(path)
        outputs = model(input_data)[0]["token_ids"]
        restored_outputs = restored_model(input_data)[0]["token_ids"]
        self.assertAllEqual(outputs, restored_outputs)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""DistilBERT model preset configurations."""

backbone_presets = {
    "distil_bert_base_en_uncased": {
        "metadata": {
            "description": (
                "6-layer DistilBERT model where all input is lowercased. "
                "Trained on English Wikipedia + BooksCorpus using BERT as the "
                "teacher model."
            ),
            "params": 66362880,
            "official_name": "DistilBERT",
            "path": "distil_bert",
        },
        "config": {
            "vocabulary_size": 30522,
            "num_layers": 6,
            "num_heads": 12,
            "hidden_dim": 768,
            "intermediate_dim": 3072,
            "dropout": 0.1,
            "max_sequence_length": 512,
        },
        "preprocessor_config": {
            "lowercase": True,
        },
        "weights_url": "https://storage.googleapis.com/keras-nlp/models/distil_bert_base_en_uncased/v1/model.h5",
        "weights_hash": "6625a649572e74086d74c46b8d0b0da3",
        "vocabulary_url": "https://storage.googleapis.com/keras-nlp/models/distil_bert_base_en_uncased/v1/vocab.txt",
        "vocabulary_hash": "64800d5d8528ce344256daf115d4965e",
    },
    "distil_bert_base_en": {
        "metadata": {
            "description": (
                "6-layer DistilBERT model where case is maintained. "
                "Trained on English Wikipedia + BooksCorpus using BERT as the "
                "teacher model."
            ),
            "params": 65190912,
            "official_name": "DistilBERT",
            "path": "distil_bert",
        },
        "config": {
            "vocabulary_size": 28996,
            "num_layers": 6,
            "num_heads": 12,
            "hidden_dim": 768,
            "intermediate_dim": 3072,
            "dropout": 0.1,
            "max_sequence_length": 512,
        },
        "preprocessor_config": {
            "lowercase": False,
        },
        "weights_url": "https://storage.googleapis.com/keras-nlp/models/distil_bert_base_en/v1/model.h5",
        "weights_hash": "fa36aa6865978efbf85a5c8264e5eb57",
        "vocabulary_url": "https://storage.googleapis.com/keras-nlp/models/distil_bert_base_en/v1/vocab.txt",
        "vocabulary_hash": "bb6ca9b42e790e5cd986bbb16444d0e0",
    },
    "distil_bert_base_multi": {
        "metadata": {
            "description": (
                "6-layer DistilBERT model where case is maintained. Trained on Wikipedias of 104 languages"
            ),
            "params": 134734080,
            "official_name": "DistilBERT",
            "path": "distil_bert",
        },
        "config": {
            "vocabulary_size": 119547,
            "num_layers": 6,
            "num_heads": 12,
            "hidden_dim": 768,
            "intermediate_dim": 3072,
            "dropout": 0.1,
            "max_sequence_length": 512,
        },
        "preprocessor_config": {
            "lowercase": False,
        },
        "weights_url": "https://storage.googleapis.com/keras-nlp/models/distil_bert_base_multi/v1/model.h5",
        "weights_hash": "c0f11095e2a6455bd3b1a6d14800a7fa",
        "vocabulary_url": "https://storage.googleapis.com/keras-nlp/models/distil_bert_base_multi/v1/vocab.txt",
        "vocabulary_hash": "d9d865138d17f1958502ed060ecfeeb6",
    },
}



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import sys

import tensorflow as tf
from absl import app
from absl import flags
from absl import logging
from tensorflow import keras

import keras_nlp
from examples.bert_pretraining.bert_config import MODEL_CONFIGS
from examples.bert_pretraining.bert_config import PREPROCESSING_CONFIG
from examples.bert_pretraining.bert_config import TRAINING_CONFIG

FLAGS = flags.FLAGS

flags.DEFINE_string(
    "input_directory",
    None,
    "The directory of training data. It can be a local disk path, or the URL "
    "of Google cloud storage bucket.",
)

flags.DEFINE_string(
    "saved_model_output",
    None,
    "Output directory to save the model to.",
)

flags.DEFINE_string(
    "checkpoint_save_directory",
    None,
    "Output directory to save checkpoints to.",
)

flags.DEFINE_bool(
    "skip_restore",
    False,
    "Skip restoring from checkpoint if True",
)

flags.DEFINE_bool(
    "tpu_name",
    None,
    "The TPU to connect to. If None, TPU will not be used.",
)

flags.DEFINE_bool(
    "enable_cloud_logging",
    False,
    "If True, the script will use cloud logging.",
)

flags.DEFINE_string(
    "tensorboard_log_path",
    None,
    "The path to save tensorboard log to.",
)

flags.DEFINE_string(
    "model_size",
    "tiny",
    "One of: tiny, mini, small, medium, base, or large.",
)

flags.DEFINE_string(
    "vocab_file",
    None,
    "The vocabulary file for tokenization.",
)

flags.DEFINE_integer(
    "num_train_steps",
    None,
    "Override the pre-configured number of train steps..",
)


class MaskedLMHead(keras.layers.Layer):
    """Masked language model network head for BERT.

    This layer implements a masked language model based on the provided
    transformer based encoder. It assumes that the encoder network being passed
    has a "get_embedding_table()" method.

    Example:
    ```python
    encoder = keras_nlp.models.BertBackbone(
        vocabulary_size=30552,
        num_layers=12,
        num_heads=12,
        hidden_dim=768,
        intermediate_dim=3072,
        max_sequence_length=12,
    )
    lm_layer = MaskedLMHead(embedding_table=encoder.get_embedding_table())
    ```

    Args:
        embedding_table: The embedding table from encoder network.
        intermediate_activation: The activation, if any, for the inner dense
            layer.
        initializer: The initializer for the dense layer. Defaults to a Glorot
            uniform initializer.
        output: The output style for this layer. Can be either 'logits' or
            'predictions'.
    """

    def __init__(
        self,
        embedding_table,
        intermediate_activation="gelu",
        initializer="glorot_uniform",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.embedding_table = embedding_table
        self.intermediate_activation = keras.activations.get(
            intermediate_activation
        )
        self.initializer = initializer

    def build(self, input_shape):
        self._vocab_size, hidden_dim = self.embedding_table.shape
        self.dense = keras.layers.Dense(
            hidden_dim,
            activation=self.intermediate_activation,
            kernel_initializer=self.initializer,
            name="transform/dense",
        )
        self.layer_norm = keras.layers.LayerNormalization(
            axis=-1, epsilon=1e-12, name="transform/LayerNorm"
        )
        self.bias = self.add_weight(
            "output_bias/bias",
            shape=(self._vocab_size,),
            initializer="zeros",
            trainable=True,
        )

        super().build(input_shape)

    def call(self, sequence_data, masked_positions):
        masked_lm_input = self._gather_indexes(sequence_data, masked_positions)
        lm_data = self.dense(masked_lm_input)
        lm_data = self.layer_norm(lm_data)
        lm_data = tf.matmul(lm_data, self.embedding_table, transpose_b=True)
        logits = tf.nn.bias_add(lm_data, self.bias)
        masked_positions_length = (
            masked_positions.shape.as_list()[1] or tf.shape(masked_positions)[1]
        )
        return tf.reshape(
            logits, [-1, masked_positions_length, self._vocab_size]
        )

    def _gather_indexes(self, sequence_tensor, positions):
        """Gathers the vectors at the specific positions, for performance.

        Args:
            sequence_tensor: Sequence output of shape
                (`batch_size`, `seq_length`, `hidden_dim`) where `hidden_dim`
                is number of hidden units.
            positions: Positions ids of tokens in sequence to mask for
                pretraining of with dimension (batch_size, num_predictions)
                where `num_predictions` is maximum number of tokens to mask out
                and predict per each sequence.

        Returns:
            Masked out sequence tensor of shape (batch_size * num_predictions,
            `hidden_dim`).
        """
        sequence_shape = tf.shape(sequence_tensor)
        batch_size, seq_length = sequence_shape[0], sequence_shape[1]
        width = sequence_tensor.shape.as_list()[2] or sequence_shape[2]

        flat_offsets = tf.reshape(
            tf.range(0, batch_size, dtype=tf.int32) * seq_length, [-1, 1]
        )
        flat_positions = tf.reshape(positions + flat_offsets, [-1])
        flat_sequence_tensor = tf.reshape(
            sequence_tensor, [batch_size * seq_length, width]
        )
        output_tensor = tf.gather(flat_sequence_tensor, flat_positions)

        return output_tensor


class BertPretrainingModel(keras.Model):
    """MaskedLM + NSP model with Bert encoder."""

    def __init__(self, encoder, **kwargs):
        super().__init__(**kwargs)
        self.encoder = encoder
        # TODO(jbischof): replace with keras_nlp.layers.MaskedLMHead (Issue #166)
        self.masked_lm_head = MaskedLMHead(
            embedding_table=encoder.token_embedding.embeddings,
            initializer=keras.initializers.TruncatedNormal(stddev=0.02),
            name="mlm_layer",
        )
        self.next_sentence_head = keras.layers.Dense(
            encoder.num_segments,
            kernel_initializer=keras.initializers.TruncatedNormal(stddev=0.02),
            name="nsp_layer",
        )

    def call(self, data):
        encoder_output = self.encoder(
            {
                "token_ids": data["token_ids"],
                "segment_ids": data["segment_ids"],
                "padding_mask": data["padding_mask"],
            }
        )
        sequence_output, pooled_output = (
            encoder_output["sequence_output"],
            encoder_output["pooled_output"],
        )
        lm_preds = self.masked_lm_head(
            sequence_output, data["masked_lm_positions"]
        )
        nsp_preds = self.next_sentence_head(pooled_output)
        return {"mlm": lm_preds, "nsp": nsp_preds}


class LinearDecayWithWarmup(keras.optimizers.schedules.LearningRateSchedule):
    """
    A learning rate schedule with linear warmup and decay.

    This schedule implements a linear warmup for the first `num_warmup_steps`
    and a linear ramp down until `num_train_steps`.
    """

    def __init__(self, learning_rate, num_warmup_steps, num_train_steps):
        self.learning_rate = learning_rate
        self.warmup_steps = num_warmup_steps
        self.train_steps = num_train_steps

    def __call__(self, step):
        peak_lr = tf.cast(self.learning_rate, dtype=tf.float32)
        warmup = tf.cast(self.warmup_steps, dtype=tf.float32)
        training = tf.cast(self.train_steps, dtype=tf.float32)
        step = tf.cast(step, dtype=tf.float32)

        is_warmup = step < warmup

        # Linear Warmup will be implemented if current step is less than
        # `num_warmup_steps` else Linear Decay will be implemented.
        return tf.cond(
            is_warmup,
            lambda: peak_lr * (step / warmup),
            lambda: tf.math.maximum(
                0.0, peak_lr * (training - step) / (training - warmup)
            ),
        )

    def get_config(self):
        return {
            "learning_rate": self.learning_rate,
            "num_warmup_steps": self.warmup_steps,
            "num_train_steps": self.train_steps,
        }


def decode_record(record):
    """Decodes a record to a TensorFlow example."""
    seq_length = PREPROCESSING_CONFIG["max_seq_length"]
    lm_length = PREPROCESSING_CONFIG["max_predictions_per_seq"]
    name_to_features = {
        "token_ids": tf.io.FixedLenFeature([seq_length], tf.int64),
        "padding_mask": tf.io.FixedLenFeature([seq_length], tf.int64),
        "segment_ids": tf.io.FixedLenFeature([seq_length], tf.int64),
        "masked_lm_positions": tf.io.FixedLenFeature([lm_length], tf.int64),
        "masked_lm_ids": tf.io.FixedLenFeature([lm_length], tf.int64),
        "masked_lm_weights": tf.io.FixedLenFeature([lm_length], tf.float32),
        "next_sentence_labels": tf.io.FixedLenFeature([1], tf.int64),
    }
    # tf.Example only supports tf.int64, but the TPU only supports tf.int32.
    # So cast all int64 to int32.
    example = tf.io.parse_single_example(record, name_to_features)
    for name in list(example.keys()):
        value = example[name]
        if value.dtype == tf.int64:
            value = tf.cast(value, tf.int32)
        example[name] = value

    inputs = {
        "token_ids": example["token_ids"],
        "padding_mask": example["padding_mask"],
        "segment_ids": example["segment_ids"],
        "masked_lm_positions": example["masked_lm_positions"],
    }
    labels = {
        "mlm": example["masked_lm_ids"],
        "nsp": example["next_sentence_labels"],
    }
    sample_weights = {"mlm": example["masked_lm_weights"], "nsp": tf.ones((1,))}
    sample = (inputs, labels, sample_weights)
    return sample


def get_checkpoint_callback():
    if tf.io.gfile.exists(FLAGS.checkpoint_save_directory):
        if not tf.io.gfile.isdir(FLAGS.checkpoint_save_directory):
            raise ValueError(
                "`checkpoint_save_directory` should be a directory, "
                f"but {FLAGS.checkpoint_save_directory} is not a "
                "directory. Please set `checkpoint_save_directory` as "
                "a directory."
            )

        elif FLAGS.skip_restore:
            # Clear up the directory if users want to skip restoring.
            tf.io.gfile.rmtree(FLAGS.checkpoint_save_directory)
    checkpoint_path = FLAGS.checkpoint_save_directory
    return keras.callbacks.BackupAndRestore(
        backup_dir=checkpoint_path,
    )


def get_tensorboard_callback():
    timestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    log_dir = FLAGS.tensorboard_log_path + timestamp
    return keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1)


def main(_):
    if FLAGS.enable_cloud_logging:
        # If the job is on cloud, we will use cloud logging.
        import google.cloud.logging

        keras.utils.disable_interactive_logging()
        client = google.cloud.logging.Client()
        client.setup_logging()

    logging.info(f"Reading input data from {FLAGS.input_directory}")
    if not tf.io.gfile.isdir(FLAGS.input_directory):
        raise ValueError(
            "`input_directory` should be a directory, "
            f"but {FLAGS.input_directory} is not a directory. Please "
            "set `input_directory` flag as a directory."
        )
    files = tf.io.gfile.listdir(FLAGS.input_directory)
    input_filenames = [FLAGS.input_directory + "/" + file for file in files]

    if not input_filenames:
        logging.info("No input files found. Check `input_directory` flag.")
        sys.exit(1)

    vocab = []
    with tf.io.gfile.GFile(FLAGS.vocab_file) as vocab_file:
        for line in vocab_file:
            vocab.append(line.strip())

    model_config = MODEL_CONFIGS[FLAGS.model_size]

    if FLAGS.tpu_name is None:
        # Use default strategy if not using TPU.
        strategy = tf.distribute.get_strategy()
    else:
        # Connect to TPU and create TPU strategy.
        resolver = tf.distribute.cluster_resolver.TPUClusterResolver.connect(
            tpu=FLAGS.tpu_name
        )
        strategy = tf.distribute.TPUStrategy(resolver)

    # Decode and batch data.
    dataset = tf.data.TFRecordDataset(input_filenames)
    dataset = dataset.map(
        lambda record: decode_record(record),
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )
    dataset = dataset.batch(TRAINING_CONFIG["batch_size"], drop_remainder=True)
    dataset = dataset.repeat()

    with strategy.scope():
        # Create a Bert model the input config.
        encoder = keras_nlp.models.BertBackbone(
            vocabulary_size=len(vocab), **model_config
        )
        # Make sure model has been called.
        encoder(encoder.inputs)
        encoder.summary()

        # Allow overriding train steps from the command line for quick testing.
        if FLAGS.num_train_steps is not None:
            num_train_steps = FLAGS.num_train_steps
        else:
            num_train_steps = TRAINING_CONFIG["num_train_steps"]
        num_warmup_steps = int(
            num_train_steps * TRAINING_CONFIG["warmup_percentage"]
        )
        learning_rate_schedule = LinearDecayWithWarmup(
            learning_rate=TRAINING_CONFIG["learning_rate"],
            num_warmup_steps=num_warmup_steps,
            num_train_steps=num_train_steps,
        )
        optimizer = keras.optimizers.Adam(learning_rate=learning_rate_schedule)

        lm_loss = keras.losses.SparseCategoricalCrossentropy(
            from_logits=True,
            name="lm_loss",
        )
        nsp_loss = keras.losses.SparseCategoricalCrossentropy(
            from_logits=True,
            name="nsp_loss",
        )

        lm_accuracy = keras.metrics.SparseCategoricalAccuracy(name="accuracy")
        nsp_accuracy = keras.metrics.SparseCategoricalAccuracy(name="accuracy")

        pretraining_model = BertPretrainingModel(encoder)
        pretraining_model.compile(
            optimizer=optimizer,
            loss={"mlm": lm_loss, "nsp": nsp_loss},
            weighted_metrics={"mlm": lm_accuracy, "nsp": nsp_accuracy},
        )

    epochs = TRAINING_CONFIG["epochs"]
    steps_per_epoch = num_train_steps // epochs

    callbacks = []
    if FLAGS.checkpoint_save_directory:
        callbacks.append(get_checkpoint_callback())
    if FLAGS.tensorboard_log_path:
        callbacks.append(get_tensorboard_callback())

    pretraining_model.fit(
        dataset,
        epochs=epochs,
        steps_per_epoch=steps_per_epoch,
        callbacks=callbacks,
    )

    model_path = FLAGS.saved_model_output
    logging.info(f"Saving to {FLAGS.saved_model_output}")
    encoder.save(model_path)


if __name__ == "__main__":
    flags.mark_flag_as_required("input_directory")
    flags.mark_flag_as_required("vocab_file")
    flags.mark_flag_as_required("saved_model_output")
    app.run(main)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import shutil

import numpy as np
import tensorflow as tf
import transformers
from absl import app
from absl import flags

import keras_nlp
from tools.checkpoint_conversion.checkpoint_conversion_utils import (
    get_md5_checksum,
)

PRESET_MAP = {
    "bart_base_en": "facebook/bart-base",
    "bart_large_en": "facebook/bart-large",
}

FLAGS = flags.FLAGS
flags.DEFINE_string(
    "preset", None, f'Must be one of {",".join(PRESET_MAP.keys())}'
)


def convert_checkpoints(hf_model):
    print("\n-> Convert original weights to KerasNLP format.")

    print("\n-> Load KerasNLP model.")
    keras_nlp_model = keras_nlp.models.BartBackbone.from_preset(
        FLAGS.preset, load_weights=False
    )

    hf_wts = hf_model.state_dict()
    print("Original weights:")
    print(list(hf_wts.keys()))

    hidden_dim = keras_nlp_model.hidden_dim
    num_heads = keras_nlp_model.num_heads

    # Token embedding weights shared by encoder and decoder.
    keras_nlp_model.get_layer("token_embedding").embeddings.assign(
        hf_wts["shared.weight"]
    )

    # Encoder weights.
    keras_nlp_model.get_layer(
        "encoder_position_embedding"
    ).position_embeddings.assign(hf_wts["encoder.embed_positions.weight"][2:])

    keras_nlp_model.get_layer("encoder_embeddings_layer_norm").gamma.assign(
        hf_wts["encoder.layernorm_embedding.weight"]
    )
    keras_nlp_model.get_layer("encoder_embeddings_layer_norm").beta.assign(
        hf_wts["encoder.layernorm_embedding.bias"]
    )

    for i in range(keras_nlp_model.num_layers):
        # Self-attention.
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layer._query_dense.kernel.assign(
            hf_wts[f"encoder.layers.{i}.self_attn.q_proj.weight"]
            .transpose(1, 0)
            .reshape((hidden_dim, num_heads, -1))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layer._query_dense.bias.assign(
            hf_wts[f"encoder.layers.{i}.self_attn.q_proj.bias"]
            .reshape((num_heads, -1))
            .numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layer._key_dense.kernel.assign(
            hf_wts[f"encoder.layers.{i}.self_attn.k_proj.weight"]
            .transpose(1, 0)
            .reshape((hidden_dim, num_heads, -1))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layer._key_dense.bias.assign(
            hf_wts[f"encoder.layers.{i}.self_attn.k_proj.bias"]
            .reshape((num_heads, -1))
            .numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layer._value_dense.kernel.assign(
            hf_wts[f"encoder.layers.{i}.self_attn.v_proj.weight"]
            .transpose(1, 0)
            .reshape((hidden_dim, num_heads, -1))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layer._value_dense.bias.assign(
            hf_wts[f"encoder.layers.{i}.self_attn.v_proj.bias"]
            .reshape((num_heads, -1))
            .numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layer._output_dense.kernel.assign(
            hf_wts[f"encoder.layers.{i}.self_attn.out_proj.weight"]
            .transpose(1, 0)
            .reshape((num_heads, -1, hidden_dim))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layer._output_dense.bias.assign(
            hf_wts[f"encoder.layers.{i}.self_attn.out_proj.bias"].numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layernorm.gamma.assign(
            hf_wts[f"encoder.layers.{i}.self_attn_layer_norm.weight"].numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._self_attention_layernorm.beta.assign(
            hf_wts[f"encoder.layers.{i}.self_attn_layer_norm.bias"].numpy()
        )

        # Post self-attention layers.
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._feedforward_intermediate_dense.kernel.assign(
            hf_wts[f"encoder.layers.{i}.fc1.weight"].transpose(1, 0).numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._feedforward_intermediate_dense.bias.assign(
            hf_wts[f"encoder.layers.{i}.fc1.bias"].numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._feedforward_output_dense.kernel.assign(
            hf_wts[f"encoder.layers.{i}.fc2.weight"].transpose(1, 0).numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._feedforward_output_dense.bias.assign(
            hf_wts[f"encoder.layers.{i}.fc2.bias"].numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._feedforward_layernorm.gamma.assign(
            hf_wts[f"encoder.layers.{i}.final_layer_norm.weight"].numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_encoder_layer_{i}"
        )._feedforward_layernorm.beta.assign(
            hf_wts[f"encoder.layers.{i}.final_layer_norm.bias"].numpy()
        )

    # Decoder weights.

    keras_nlp_model.get_layer(
        "decoder_position_embedding"
    ).position_embeddings.assign(hf_wts["decoder.embed_positions.weight"][2:])

    keras_nlp_model.get_layer("decoder_embeddings_layer_norm").gamma.assign(
        hf_wts["decoder.layernorm_embedding.weight"]
    )
    keras_nlp_model.get_layer("decoder_embeddings_layer_norm").beta.assign(
        hf_wts["decoder.layernorm_embedding.bias"]
    )

    for i in range(keras_nlp_model.num_layers):
        # Self-attention.
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layer._query_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.self_attn.q_proj.weight"]
            .transpose(1, 0)
            .reshape((hidden_dim, num_heads, -1))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layer._query_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.self_attn.q_proj.bias"]
            .reshape((num_heads, -1))
            .numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layer._key_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.self_attn.k_proj.weight"]
            .transpose(1, 0)
            .reshape((hidden_dim, num_heads, -1))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layer._key_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.self_attn.k_proj.bias"]
            .reshape((num_heads, -1))
            .numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layer._value_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.self_attn.v_proj.weight"]
            .transpose(1, 0)
            .reshape((hidden_dim, num_heads, -1))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layer._value_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.self_attn.v_proj.bias"]
            .reshape((num_heads, -1))
            .numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layer._output_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.self_attn.out_proj.weight"]
            .transpose(1, 0)
            .reshape((num_heads, -1, hidden_dim))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layer._output_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.self_attn.out_proj.bias"].numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layernorm.gamma.assign(
            hf_wts[f"decoder.layers.{i}.self_attn_layer_norm.weight"].numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._self_attention_layernorm.beta.assign(
            hf_wts[f"decoder.layers.{i}.self_attn_layer_norm.bias"].numpy()
        )

        # Cross-attention.
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layer._query_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn.q_proj.weight"]
            .transpose(1, 0)
            .reshape((hidden_dim, num_heads, -1))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layer._query_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn.q_proj.bias"]
            .reshape((num_heads, -1))
            .numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layer._key_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn.k_proj.weight"]
            .transpose(1, 0)
            .reshape((hidden_dim, num_heads, -1))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layer._key_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn.k_proj.bias"]
            .reshape((num_heads, -1))
            .numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layer._value_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn.v_proj.weight"]
            .transpose(1, 0)
            .reshape((hidden_dim, num_heads, -1))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layer._value_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn.v_proj.bias"]
            .reshape((num_heads, -1))
            .numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layer._output_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn.out_proj.weight"]
            .transpose(1, 0)
            .reshape((num_heads, -1, hidden_dim))
            .numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layer._output_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn.out_proj.bias"].numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layernorm.gamma.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn_layer_norm.weight"].numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._cross_attention_layernorm.beta.assign(
            hf_wts[f"decoder.layers.{i}.encoder_attn_layer_norm.bias"].numpy()
        )

        # Post self-attention and cross-attention layers.
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._feedforward_intermediate_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.fc1.weight"].transpose(1, 0).numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._feedforward_intermediate_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.fc1.bias"].numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._feedforward_output_dense.kernel.assign(
            hf_wts[f"decoder.layers.{i}.fc2.weight"].transpose(1, 0).numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._feedforward_output_dense.bias.assign(
            hf_wts[f"decoder.layers.{i}.fc2.bias"].numpy()
        )

        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._feedforward_layernorm.gamma.assign(
            hf_wts[f"decoder.layers.{i}.final_layer_norm.weight"].numpy()
        )
        keras_nlp_model.get_layer(
            f"transformer_decoder_layer_{i}"
        )._feedforward_layernorm.beta.assign(
            hf_wts[f"decoder.layers.{i}.final_layer_norm.bias"].numpy()
        )

    # Save the model.
    print("\n-> Save KerasNLP model weights.")
    keras_nlp_model.save_weights(os.path.join(FLAGS.preset, "model.h5"))

    return keras_nlp_model


def extract_vocab(hf_tokenizer):
    vocabulary_path = os.path.join(FLAGS.preset, "vocab.json")
    merges_path = os.path.join(FLAGS.preset, "merges.txt")
    print(f"\n-> Save KerasNLP vocab to `{vocabulary_path}`.")
    print(f"-> Save KerasNLP merges to `{merges_path}`.")

    # Huggingface has a save_vocabulary function but it's not byte-for-byte
    # with the source. Instead copy the original downloaded file directly.
    shutil.copyfile(
        transformers.utils.hub.get_file_from_repo(
            hf_tokenizer.name_or_path, "vocab.json"
        ),
        vocabulary_path,
    )
    shutil.copyfile(
        transformers.utils.hub.get_file_from_repo(
            hf_tokenizer.name_or_path, "merges.txt"
        ),
        merges_path,
    )

    keras_nlp_tokenizer = keras_nlp.models.BartTokenizer(
        vocabulary=vocabulary_path, merges=merges_path
    )

    print("-> Print MD5 checksum of the vocab files.")
    print(f"`{vocabulary_path}` md5sum: ", get_md5_checksum(vocabulary_path))
    print(f"`{merges_path}` md5sum: ", get_md5_checksum(merges_path))

    return keras_nlp_tokenizer


def check_output(
    keras_nlp_tokenizer,
    keras_nlp_model,
    hf_tokenizer,
    hf_model,
):
    print("\n-> Check the outputs.")
    enc_sample_text = [
        "cricket is awesome, easily the best sport in the world!"
    ]
    dec_sample_text = [
        "football is good too, but nowhere near as good as cricket."
    ]

    # KerasNLP
    keras_nlp_enc_token_ids = keras_nlp_tokenizer(
        tf.constant(enc_sample_text)
    ).to_tensor()
    keras_nlp_enc_token_ids = tf.concat(
        [
            tf.constant([[keras_nlp_tokenizer.start_token_id]]),
            keras_nlp_enc_token_ids,
            tf.constant([[keras_nlp_tokenizer.end_token_id]]),
        ],
        axis=-1,
    )
    keras_nlp_dec_token_ids = keras_nlp_tokenizer(
        tf.constant(dec_sample_text)
    ).to_tensor()
    keras_nlp_dec_token_ids = tf.concat(
        [
            tf.constant([[keras_nlp_tokenizer.start_token_id]]),
            keras_nlp_dec_token_ids,
            tf.constant([[keras_nlp_tokenizer.end_token_id]]),
        ],
        axis=-1,
    )
    keras_nlp_inputs = {
        "encoder_token_ids": keras_nlp_enc_token_ids,
        "encoder_padding_mask": keras_nlp_enc_token_ids
        != keras_nlp_tokenizer.pad_token_id,
        "decoder_token_ids": keras_nlp_dec_token_ids,
        "decoder_padding_mask": keras_nlp_dec_token_ids
        != keras_nlp_tokenizer.pad_token_id,
    }
    keras_nlp_output = keras_nlp_model.predict(keras_nlp_inputs)

    # HF
    hf_enc_inputs = hf_tokenizer(enc_sample_text, return_tensors="pt")
    hf_dec_inputs = hf_tokenizer(dec_sample_text, return_tensors="pt")

    hf_output = hf_model(
        **hf_enc_inputs,
        decoder_input_ids=hf_dec_inputs["input_ids"],
        decoder_attention_mask=hf_dec_inputs["attention_mask"],
    )

    print("Encoder Outputs:")
    print(
        "KerasNLP output:",
        keras_nlp_output["encoder_sequence_output"][0, 0, :10],
    )
    print("HF output:", hf_output.encoder_last_hidden_state[0, 0, :10])
    print(
        "Difference:",
        np.mean(
            keras_nlp_output["encoder_sequence_output"]
            - hf_output.encoder_last_hidden_state.detach().numpy()
        ),
    )

    print("Decoder Outputs:")
    print(
        "KerasNLP output:",
        keras_nlp_output["decoder_sequence_output"][0, 0, :10],
    )
    print("HF output:", hf_output.last_hidden_state[0, 0, :10])
    print(
        "Difference:",
        np.mean(
            keras_nlp_output["decoder_sequence_output"]
            - hf_output.last_hidden_state.detach().numpy()
        ),
    )

    # Show the MD5 checksum of the model weights.
    print(
        "Model md5sum: ",
        get_md5_checksum(os.path.join(FLAGS.preset, "model.h5")),
    )


def main(_):
    os.makedirs(FLAGS.preset)

    hf_model_name = PRESET_MAP[FLAGS.preset]

    print("\n-> Load HF model and HF tokenizer.")
    hf_model = transformers.AutoModel.from_pretrained(hf_model_name)
    hf_model.eval()
    hf_tokenizer = transformers.AutoTokenizer.from_pretrained(hf_model_name)

    keras_nlp_model = convert_checkpoints(hf_model)
    print("\n -> Load KerasNLP tokenizer.")
    keras_nlp_tokenizer = extract_vocab(hf_tokenizer)

    check_output(
        keras_nlp_tokenizer,
        keras_nlp_model,
        hf_tokenizer,
        hf_model,
    )


if __name__ == "__main__":
    flags.mark_flag_as_required("preset")
    app.run(main)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO(jbischof): remove in favor of presets with load_weights=False
MODEL_CONFIGS = {
    "tiny": {
        "num_layers": 2,
        "hidden_dim": 128,
        "dropout": 0.1,
        "num_heads": 2,
        "intermediate_dim": 512,
    },
    "mini": {
        "num_layers": 4,
        "hidden_dim": 256,
        "dropout": 0.1,
        "num_heads": 4,
        "intermediate_dim": 1024,
    },
    "small": {
        "num_layers": 4,
        "hidden_dim": 512,
        "dropout": 0.1,
        "num_heads": 8,
        "intermediate_dim": 2048,
    },
    "medium": {
        "num_layers": 8,
        "hidden_dim": 512,
        "dropout": 0.1,
        "num_heads": 8,
        "intermediate_dim": 2048,
    },
    "base": {
        "num_layers": 12,
        "hidden_dim": 768,
        "dropout": 0.1,
        "num_heads": 12,
        "intermediate_dim": 3072,
    },
    "large": {
        "num_layers": 24,
        "hidden_dim": 1024,
        "dropout": 0.1,
        "num_heads": 16,
        "intermediate_dim": 4096,
    },
}

# Currently we have the same set of training parameters for all configurations.
# We should see if we need to split this for different architecture sizes.

PREPROCESSING_CONFIG = {
    "max_seq_length": 512,
    "max_predictions_per_seq": 76,
    "dupe_factor": 10,
    "masked_lm_prob": 0.15,
    "short_seq_prob": 0.1,
}

TRAINING_CONFIG = {
    "batch_size": 256,
    "epochs": 10,
    "learning_rate": 1e-4,
    "num_train_steps": 1_000_000,
    # Percentage of training steps used for learning rate warmup.
    "warmup_percentage": 0.1,
}



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for position embedding layer."""

import os

import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_nlp.layers import position_embedding


def custom_init(shape, dtype=None):
    count = 1
    for length in shape:
        count *= length
    return tf.reshape(tf.range(count, dtype=dtype), shape)


class PositionEmbeddingLayerTest(tf.test.TestCase, parameterized.TestCase):
    def test_static_layer_output_shape(self):
        # Create a 3-dimensional input (the first dimension is implicit).
        sequence_length = 21
        feature_size = 30
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=sequence_length
        )
        input_tensor = keras.Input(shape=(sequence_length, feature_size))
        output_tensor = test_layer(input_tensor)

        # When using static position embedding shapes, the output is expected
        # to be the same as the input shape in all dimensions save batch.
        expected_output_shape = [None, sequence_length, feature_size]
        self.assertEqual(expected_output_shape, output_tensor.shape.as_list())
        # The default output dtype for this layer should be tf.float32.
        self.assertEqual(tf.float32, output_tensor.dtype)

    def test_more_than_3_dimensions_static(self):
        # Create a 4-dimensional input (the first dimension is implicit).
        sequence_length = 21
        feature_size = 30
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=sequence_length
        )
        input_tensor = keras.Input(
            shape=(feature_size, sequence_length, feature_size)
        )
        output_tensor = test_layer(input_tensor)

        # When using static position embedding shapes, the output is expected
        # to be the same as the input shape in all dimensions save batch.
        expected_output_shape = [
            None,
            feature_size,
            sequence_length,
            feature_size,
        ]
        self.assertEqual(expected_output_shape, output_tensor.shape.as_list())
        # The default output dtype for this layer should be tf.float32.
        self.assertEqual(tf.float32, output_tensor.dtype)

    def test_float16_dtype(self):
        # Create a 3-dimensional input (the first dimension is implicit).
        sequence_length = 21
        feature_size = 30
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=sequence_length, dtype="float16"
        )
        input_tensor = keras.Input(shape=(sequence_length, feature_size))
        output_tensor = test_layer(input_tensor)

        # When using static position embedding shapes, the output is expected
        # to be the same as the input shape in all dimensions save batch.
        expected_output_shape = [None, sequence_length, feature_size]
        self.assertEqual(expected_output_shape, output_tensor.shape.as_list())
        # The default output dtype for this layer should be tf.float32.
        self.assertEqual(tf.float16, output_tensor.dtype)

    def test_dynamic_layer_output_shape(self):
        max_sequence_length = 40
        feature_size = 30
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=max_sequence_length
        )
        # Create a 3-dimensional input (the first dimension is implicit).
        input_tensor = keras.Input(shape=(None, feature_size))
        output_tensor = test_layer(input_tensor)

        # When using dynamic position embedding shapes, the output is expected
        # to be the same as the input shape in all dimensions - but may be None
        # if the input shape is None there.
        expected_output_shape = [None, None, feature_size]
        self.assertEqual(expected_output_shape, output_tensor.shape.as_list())

    def test_more_than_3_dimensions_dynamic(self):
        max_sequence_length = 60
        feature_size = 30
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=max_sequence_length
        )
        # Create a 4-dimensional input (the first dimension is implicit).
        input_tensor = keras.Input(shape=(None, None, feature_size))
        output_tensor = test_layer(input_tensor)

        # When using dynamic position embedding shapes, the output is expected
        # to be the same as the input shape in all dimensions - but may be None
        # if the input shape is None there.
        expected_output_shape = [None, None, None, feature_size]
        self.assertEqual(expected_output_shape, output_tensor.shape.as_list())

    def test_dynamic_layer_slicing(self):
        max_sequence_length = 40
        feature_size = 30
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=max_sequence_length
        )
        # Create a 3-dimensional input (the first dimension is implicit).
        input_tensor = keras.Input(shape=(None, feature_size))
        output_tensor = test_layer(input_tensor)

        model = keras.Model(input_tensor, output_tensor)

        # Create input data that is shorter than max_sequence_length, which
        # should trigger a down-slice.
        input_length = 17
        # Note: This test explicitly uses a batch size of 1. This is to get
        # around Keras' restriction on Model invocations: inputs are expected to
        # have the same batch cardinality as outputs. In practice, this layer
        # should be used inside a model, where it can be projected when added to
        # another tensor.
        input_data = tf.ones(shape=[1, input_length, feature_size])
        output_data = model.predict(input_data)

        self.assertAllEqual([1, input_length, feature_size], output_data.shape)

    def test_callable_initializer(self):
        max_sequence_length = 4
        feature_size = 3
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=max_sequence_length,
            initializer=custom_init,
        )
        inputs = keras.Input(shape=(max_sequence_length, feature_size))
        outputs = test_layer(inputs)
        model = keras.Model(inputs=inputs, outputs=outputs)

        batch_size = 2
        data = tf.zeros(shape=[batch_size, max_sequence_length, feature_size])
        model(data)
        model_output = model.predict(data)
        expected_output = tf.broadcast_to(
            tf.reshape(
                tf.range(max_sequence_length * feature_size),
                [max_sequence_length, feature_size],
            ),
            [batch_size, max_sequence_length, feature_size],
        )
        self.assertAllClose(model_output, expected_output)

    def test_ragged_tensor_with_3_dimensions(self):
        max_sequence_length = 4
        feature_size = 2
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=max_sequence_length,
            initializer=custom_init,
        )
        # Create a 3-dimensional ragged input (the first dimension is implicit).
        input_tensor = keras.Input(
            shape=(None, feature_size), dtype=tf.float32, ragged=True
        )
        output_tensor = test_layer(input_tensor)
        model = keras.Model(input_tensor, output_tensor)

        input_data = tf.ragged.constant(
            [
                [[1.0, 1.0], [1.0, 1.0]],
                [],
                [[1.0, 1.0], [1.0, 1.0], [1.0, 1.0]],
                [[1.0, 1.0]],
            ],
            ragged_rank=1,
            inner_shape=(2,),
        )
        expected_output_data = tf.ragged.constant(
            [
                [[0.0, 1.0], [2.0, 3.0]],
                [],
                [[0.0, 1.0], [2.0, 3.0], [4.0, 5.0]],
                [[0.0, 1.0]],
            ],
            ragged_rank=1,
            inner_shape=(2,),
        )
        output_data = model.predict(input_data)
        self.assertAllClose(output_data, expected_output_data)

    def test_ragged_tensor_with_4_dimensions(self):
        max_sequence_length = 4
        feature_size = 2
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=max_sequence_length,
            initializer=custom_init,
        )
        # Create a 4-dimensional ragged input (the first dimension is implicit).
        input_tensor = keras.Input(
            shape=(None, None, feature_size), dtype=tf.float32, ragged=True
        )
        output_tensor = test_layer(input_tensor)
        model = keras.Model(input_tensor, output_tensor)

        input_data = tf.ragged.constant(
            [
                [
                    [[1.0, 1.0], [1.0, 1.0]],
                    [],
                ],
                [
                    [[1.0, 1.0], [1.0, 1.0], [1.0, 1.0]],
                    [[1.0, 1.0]],
                ],
            ],
            ragged_rank=2,
            inner_shape=(2,),
        )
        expected_output_data = tf.ragged.constant(
            [
                [
                    [[0.0, 1.0], [2.0, 3.0]],
                    [],
                ],
                [
                    [[0.0, 1.0], [2.0, 3.0], [4.0, 5.0]],
                    [[0.0, 1.0]],
                ],
            ],
            ragged_rank=2,
            inner_shape=(2,),
        )
        output_data = model.predict(input_data)
        self.assertAllClose(output_data, expected_output_data)

    def test_one_training_step(self):
        max_sequence_length = 4
        feature_size = 3
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=max_sequence_length
        )
        inputs = keras.Input(shape=(max_sequence_length, feature_size))
        outputs = test_layer(inputs)
        model = keras.Model(inputs=inputs, outputs=outputs)

        batch_size = 2
        data = tf.random.uniform(
            shape=[batch_size, max_sequence_length, feature_size]
        )
        label = tf.random.uniform(shape=[max_sequence_length, feature_size])

        loss_fn = keras.losses.MeanSquaredError()
        optimizer = keras.optimizers.Adam()
        with tf.GradientTape() as tape:
            pred = model(data)
            loss = loss_fn(label, pred)
        grad = tape.gradient(loss, model.trainable_variables)
        self.assertEqual(len(grad), 1)

        trainable_variables_before = tf.Variable(model.trainable_variables[0])
        optimizer.apply_gradients(zip(grad, model.trainable_variables))
        self.assertNotAllClose(
            trainable_variables_before, model.trainable_variables[0]
        )

    def test_get_config_and_from_config(self):
        max_sequence_length = 40
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=max_sequence_length
        )
        config = test_layer.get_config()
        expected_config_subset = {
            "sequence_length": max_sequence_length,
            "initializer": {
                "class_name": "GlorotUniform",
                "config": {"seed": None},
            },
        }
        self.assertEqual(config, {**config, **expected_config_subset})

        restored_encoder = position_embedding.PositionEmbedding.from_config(
            config,
        )
        self.assertEqual(
            restored_encoder.get_config(), {**config, **expected_config_subset}
        )

    @parameterized.named_parameters(
        ("tf_format", "tf", "model"),
        ("keras_format", "keras_v3", "model.keras"),
    )
    def test_saved_model(self, save_format, filename):
        max_sequence_length = 4
        feature_size = 6
        test_layer = position_embedding.PositionEmbedding(
            sequence_length=max_sequence_length
        )
        inputs = keras.Input(shape=(max_sequence_length, feature_size))
        outputs = test_layer(inputs)
        model = keras.Model(inputs=inputs, outputs=outputs)

        data = tf.zeros(shape=[2, max_sequence_length, feature_size])
        model(data)

        path = os.path.join(self.get_temp_dir(), filename)
        model.save(path, save_format=save_format)
        loaded_model = keras.models.load_model(path)

        model_output = model.predict(data)
        loaded_model_output = loaded_model.predict(data)
        self.assertAllClose(model_output, loaded_model_output)


if __name__ == "__main__":
    tf.test.main()



# Copyright 2021 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup script."""

import pathlib

from setuptools import find_packages
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="keras-nlp",
    description=(
        "Industry-strength Natural Language Processing extensions for Keras."
    ),
    long_description=README,
    long_description_content_type="text/markdown",
    version="0.4.1",
    url="https://github.com/keras-team/keras-nlp",
    author="Keras team",
    author_email="keras-nlp@google.com",
    license="Apache License 2.0",
    install_requires=[
        "absl-py",
        "numpy",
        "packaging",
        # Don't require tensorflow-text on MacOS, there are no binaries for ARM.
        # Also, we rely on tensorflow *transitively* through tensorflow-text.
        # This avoid a slowdown during `pip install keras-nlp` where pip would
        # download many version of both libraries to find compatible versions.
        "tensorflow-text; platform_system != 'Darwin'",
    ],
    extras_require={
        "extras": [
            "rouge-score",
            "sentencepiece",
        ],
    },
    # Supported Python versions
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
    ],
    packages=find_packages(exclude=("*_test.py",)),
)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import shutil

import numpy as np
import tensorflow as tf
import transformers
from absl import app
from absl import flags

import keras_nlp
from tools.checkpoint_conversion.checkpoint_conversion_utils import (
    get_md5_checksum,
)

PRESET_MAP = {
    "opt_125m_en": "facebook/opt-125m",
    "opt_1.3b_en": "facebook/opt-1.3b",
    "opt_2.7b_en": "facebook/opt-2.7b",
    "opt_6.7b_en": "facebook/opt-6.7b",
}

FLAGS = flags.FLAGS
flags.DEFINE_string(
    "preset", None, f'Must be one of {",".join(PRESET_MAP.keys())}'
)


def convert_weights(hf_model):
    print("\n-> Convert original weights to KerasNLP format.")

    # Load PyTorch OPT checkpoint.
    keras_nlp_model = keras_nlp.models.OPTBackbone.from_preset(
        FLAGS.preset, load_weights=False
    )

    # Token embedding.
    keras_nlp_model.get_layer("embeddings").token_embedding.embeddings.assign(
        hf_model.model.decoder.embed_tokens.weight
    )
    # Position embedding.
    keras_nlp_model.get_layer(
        "embeddings"
    ).position_embedding.position_embeddings.assign(
        hf_model.model.decoder.embed_positions.weight[2:, :]
    )

    num_heads = keras_nlp_model.num_heads
    hidden_dim = keras_nlp_model.hidden_dim

    # Transformer layers.
    for i in range(keras_nlp_model.num_layers):
        # Self-attention.
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._query_dense.kernel.assign(
            tf.reshape(
                hf_model.model.decoder.layers[i].self_attn.q_proj.kernel,
                (hidden_dim, num_heads, -1),
            )
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._query_dense.bias.assign(
            tf.reshape(
                hf_model.model.decoder.layers[i].self_attn.q_proj.bias,
                (num_heads, -1),
            )
        )

        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._key_dense.kernel.assign(
            tf.reshape(
                hf_model.model.decoder.layers[i].self_attn.k_proj.kernel,
                (hidden_dim, num_heads, -1),
            )
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._key_dense.bias.assign(
            tf.reshape(
                hf_model.model.decoder.layers[i].self_attn.k_proj.bias,
                (num_heads, -1),
            )
        )

        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._value_dense.kernel.assign(
            tf.reshape(
                hf_model.model.decoder.layers[i].self_attn.v_proj.kernel,
                (hidden_dim, num_heads, -1),
            )
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._value_dense.bias.assign(
            tf.reshape(
                hf_model.model.decoder.layers[i].self_attn.v_proj.bias,
                (num_heads, -1),
            )
        )

        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._output_dense.kernel.assign(
            tf.reshape(
                hf_model.model.decoder.layers[i].self_attn.out_proj.kernel,
                (num_heads, -1, hidden_dim),
            )
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layer._output_dense.bias.assign(
            hf_model.model.decoder.layers[i].self_attn.out_proj.bias,
        )

        # Attention LayerNorm
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layernorm.gamma.assign(
            hf_model.model.decoder.layers[i].self_attn_layer_norm.gamma
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._self_attention_layernorm.beta.assign(
            hf_model.model.decoder.layers[i].self_attn_layer_norm.beta
        )

        # Intermediate FF layer
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_intermediate_dense.kernel.assign(
            hf_model.model.decoder.layers[i].fc1.kernel
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_intermediate_dense.bias.assign(
            hf_model.model.decoder.layers[i].fc1.bias
        )

        # Output dense layer
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_output_dense.kernel.assign(
            hf_model.model.decoder.layers[i].fc2.kernel
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_output_dense.bias.assign(
            hf_model.model.decoder.layers[i].fc2.bias
        )

        # FF LayerNorm
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_layernorm.gamma.assign(
            hf_model.model.decoder.layers[i].final_layer_norm.gamma
        )
        keras_nlp_model.get_layer(
            f"transformer_layer_{i}"
        )._feedforward_layernorm.beta.assign(
            hf_model.model.decoder.layers[i].final_layer_norm.beta
        )

    # Output LayerNorm
    keras_nlp_model.get_layer("layer_norm").gamma.assign(
        hf_model.model.decoder.final_layer_norm.gamma
    )
    keras_nlp_model.get_layer("layer_norm").beta.assign(
        hf_model.model.decoder.final_layer_norm.beta
    )

    # Save the model.
    model_path = f"./{FLAGS.preset}/model.h5"
    print(f"-> Save KerasNLP model weights to `{model_path}`.")
    keras_nlp_model.save_weights(model_path)
    print("-> Print MD5 checksum of the model weights files.")
    print(f"`{model_path}` md5sum: ", get_md5_checksum(model_path))

    return keras_nlp_model


def extract_vocab(hf_tokenizer):
    vocabulary_path = f"./{FLAGS.preset}/vocab.json"
    merges_path = f"./{FLAGS.preset}/merges.txt"
    print(f"\n-> Save KerasNLP vocab to `{vocabulary_path}`.")
    print(f"-> Save KerasNLP merges to `{merges_path}`.")

    # Huggingface has a save_vocabulary function but it's not byte-for-byte
    # with the source. Instead copy the original downloaded file directly.
    shutil.copyfile(
        transformers.utils.hub.get_file_from_repo(
            hf_tokenizer.name_or_path, "vocab.json"
        ),
        vocabulary_path,
    )
    shutil.copyfile(
        transformers.utils.hub.get_file_from_repo(
            hf_tokenizer.name_or_path, "merges.txt"
        ),
        merges_path,
    )

    keras_nlp_tokenizer = keras_nlp.models.OPTTokenizer(
        vocabulary=vocabulary_path, merges=merges_path
    )

    print("-> Print MD5 checksum of the vocab files.")
    print(f"`{vocabulary_path}` md5sum: ", get_md5_checksum(vocabulary_path))
    print(f"`{merges_path}` md5sum: ", get_md5_checksum(merges_path))

    return keras_nlp_tokenizer


def check_output(
    keras_nlp_model,
    keras_nlp_tokenizer,
    hf_model,
    hf_tokenizer,
):
    print("\n-> Check the outputs.")
    input_str = ["the quick brown fox ran, galloped and jumped."]

    sequence_length = 16
    packer = keras_nlp.layers.StartEndPacker(
        sequence_length=sequence_length,
        start_value=keras_nlp_tokenizer.start_token_id,
        pad_value=keras_nlp_tokenizer.pad_token_id,
    )

    # KerasNLP
    token_ids = packer(keras_nlp_tokenizer(input_str))
    padding_mask = token_ids != keras_nlp_tokenizer.pad_token_id
    keras_nlp_inputs = {
        "token_ids": token_ids,
        "padding_mask": padding_mask,
    }
    keras_nlp_output = keras_nlp_model(keras_nlp_inputs)

    # HF
    hf_inputs = hf_tokenizer(
        input_str,
        padding="max_length",
        max_length=sequence_length,
        return_tensors="tf",
    )
    hf_output = hf_model(
        **hf_inputs, return_dict=True, output_hidden_states=True
    )

    # Compare tokenized inputs. This should be a compete match.
    print("KerasNLP inputs:", keras_nlp_inputs)
    print("HF inputs:", hf_inputs)

    # Compare outputs, this should match closely, though not exactly.
    hf_output = hf_output.last_hidden_state
    print("KerasNLP output:", keras_nlp_output[0, 0, :5])
    print("HF output:", hf_output[0, 0, :5])
    difference = keras_nlp_output - hf_output
    difference_non_padding = tf.gather_nd(difference, tf.where(padding_mask))
    print("Difference:", np.mean(difference_non_padding))


def main(_):
    hf_id = PRESET_MAP[FLAGS.preset]
    os.mkdir(f"./{FLAGS.preset}")

    print("\n-> Load HF model.")
    hf_tokenizer = transformers.AutoTokenizer.from_pretrained(hf_id)
    hf_model = transformers.TFAutoModel.from_pretrained(hf_id)

    keras_nlp_tokenizer = extract_vocab(hf_tokenizer)
    keras_nlp_model = convert_weights(hf_model)

    check_output(
        keras_nlp_model,
        keras_nlp_tokenizer,
        hf_model,
        hf_tokenizer,
    )


if __name__ == "__main__":
    flags.mark_flag_as_required("preset")
    app.run(main)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Beam sampler."""

import random

import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_nlp.samplers.beam_sampler import BeamSampler
from keras_nlp.samplers.greedy_sampler import GreedySampler


class BeamSamplerTest(tf.test.TestCase, parameterized.TestCase):
    def setUp(self):
        super().setUp()
        self.vocab_size = 10
        self.feature_size = 16

        # Create a dummy model to predict the next token.
        model = keras.Sequential(
            [
                keras.Input(shape=[None]),
                keras.layers.Embedding(
                    input_dim=self.vocab_size,
                    output_dim=self.feature_size,
                ),
                keras.layers.Dense(self.vocab_size),
                keras.layers.Softmax(),
            ]
        )

        def token_probability_fn(inputs, mask):
            return model(inputs)

        self.token_probability_fn = token_probability_fn
        self.sampler = BeamSampler(num_beams=2)

    def test_generate_with_1d_prompt(self):
        inputs = tf.constant([1])
        outputs = self.sampler(inputs, self.token_probability_fn, max_length=5)
        self.assertEqual(outputs.shape, [5])

    def test_generate_with_2d_prompt(self):
        inputs = tf.constant([[1], [1]])
        outputs = self.sampler(inputs, self.token_probability_fn, max_length=5)
        self.assertEqual(outputs.shape, [2, 5])

    def test_generate_with_list_prompt(self):
        inputs = [[1], [1]]
        outputs = self.sampler(inputs, self.token_probability_fn, max_length=5)
        self.assertEqual(outputs.shape, [2, 5])

    def test_generate_with_ragged_prompt(self):
        def token_probability_fn(inputs, mask):
            batch_size, seq_length = tf.shape(inputs)[0], tf.shape(inputs)[1]
            prob = tf.constant([[[0.0, 0.0, 0.0, 1.0]]])
            return tf.tile(prob, [batch_size, seq_length, 1])

        inputs = tf.ragged.constant([[1], [2, 1, 2]])
        outputs = self.sampler(inputs, token_probability_fn, max_length=5)
        self.assertEqual(outputs.shape, [2, 5])

    def test_one_beam_generation(self):
        for _ in range(5):
            inputs = tf.constant([random.randint(0, 9)])
            beam_sampler = BeamSampler(num_beams=1)
            greedy_sampler = GreedySampler()
            beam_output = beam_sampler(
                inputs,
                self.token_probability_fn,
                max_length=5,
            )
            greedy_output = greedy_sampler(
                inputs,
                self.token_probability_fn,
                max_length=5,
            )
            self.assertAllEqual(beam_output, greedy_output)

    @parameterized.named_parameters(
        ("xla_graph", True, False),
        ("non_xla_graph", False, False),
        ("eager", False, True),
    )
    def test_assert_generation_is_correct(self, jit_compile, run_eagerly):
        def token_probability_fn(inputs, mask):
            batch_size, seq_length = tf.shape(inputs)[0], tf.shape(inputs)[1]
            prob = tf.constant([[[0.0, 0.0, 0.0, 1.0]]])
            return tf.tile(prob, [batch_size, seq_length, 1])

        batch_size = 10
        inputs = 3 * tf.ones([batch_size, 1], dtype=tf.int32)
        max_length = 3
        for i in range(1, 5):
            sampler = BeamSampler(
                num_beams=i,
                jit_compile=jit_compile,
                run_eagerly=run_eagerly,
            )
            outputs = sampler(
                inputs,
                token_probability_fn,
                max_length=max_length,
            )
            self.assertAllEqual(
                outputs, 3 * tf.ones(shape=[batch_size, max_length])
            )

    def test_end_token_id(self):
        def token_probability_fn(inputs, mask):
            batch_size, seq_length = tf.shape(inputs)[0], tf.shape(inputs)[1]
            prob = tf.constant([[[0.0, 0.0, 0.0, 1.0]]])
            return tf.tile(prob, [batch_size, seq_length, 1])

        max_length = 4
        inputs = tf.constant([[0, 1], [1, 2]])
        outputs = self.sampler(
            inputs,
            token_probability_fn,
            max_length=max_length,
            end_token_id=2,
        )
        # end_token in prompt does not trigger truncation.
        expected_outputs = tf.ragged.constant([[0, 1, 3, 3], [1, 2, 3, 3]])
        self.assertAllEqual(outputs, expected_outputs)

        max_length = 4
        inputs = tf.constant([[0, 1], [1, 3]])
        outputs = self.sampler(
            inputs,
            token_probability_fn,
            max_length=max_length,
            end_token_id=3,
        )
        expected_outputs = tf.ragged.constant([[0, 1], [1, 3]])
        self.assertAllEqual(outputs, expected_outputs)



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for ALBERT backbone model."""

import os

import tensorflow as tf
from absl.testing import parameterized
from tensorflow import keras

from keras_nlp.models.albert.albert_backbone import AlbertBackbone


class AlbertBackboneTest(tf.test.TestCase, parameterized.TestCase):
    def setUp(self):
        self.model = AlbertBackbone(
            vocabulary_size=1000,
            num_layers=2,
            num_heads=2,
            num_groups=1,
            num_inner_repetitions=1,
            embedding_dim=16,
            hidden_dim=64,
            intermediate_dim=128,
            max_sequence_length=128,
        )
        self.batch_size = 8
        self.input_batch = {
            "token_ids": tf.ones(
                (self.batch_size, self.model.max_sequence_length), dtype="int32"
            ),
            "segment_ids": tf.ones(
                (self.batch_size, self.model.max_sequence_length), dtype="int32"
            ),
            "padding_mask": tf.ones(
                (self.batch_size, self.model.max_sequence_length), dtype="int32"
            ),
        }

        self.input_dataset = tf.data.Dataset.from_tensor_slices(
            self.input_batch
        ).batch(2)

    def test_valid_call_albert(self):
        self.model(self.input_batch)

        # Check default name passed through
        self.assertRegexpMatches(self.model.name, "albert_backbone")

    def test_variable_sequence_length_call_albert(self):
        for seq_length in (25, 50, 75):
            input_data = {
                "token_ids": tf.ones(
                    (self.batch_size, seq_length), dtype="int32"
                ),
                "segment_ids": tf.ones(
                    (self.batch_size, seq_length), dtype="int32"
                ),
                "padding_mask": tf.ones(
                    (self.batch_size, seq_length), dtype="int32"
                ),
            }
            self.model(input_data)

    @parameterized.named_parameters(
        ("jit_compile_false", False), ("jit_compile_true", True)
    )
    def test_compile(self, jit_compile):
        self.model.compile(jit_compile=jit_compile)
        self.model.predict(self.input_batch)

    @parameterized.named_parameters(
        ("jit_compile_false", False), ("jit_compile_true", True)
    )
    def test_compile_batched_ds(self, jit_compile):
        self.model.compile(jit_compile=jit_compile)
        self.model.predict(self.input_dataset)

    def test_error_for_invalid_num_groups(self):
        with self.assertRaises(ValueError):
            self.model = AlbertBackbone(
                vocabulary_size=1000,
                num_layers=3,
                num_heads=2,
                num_groups=2,
                num_inner_repetitions=1,
                embedding_dim=16,
                hidden_dim=64,
                intermediate_dim=128,
            )

    @parameterized.named_parameters(
        ("tf_format", "tf", "model"),
        ("keras_format", "keras_v3", "model.keras"),
    )
    def test_saved_model(self, save_format, filename):
        model_output = self.model(self.input_batch)
        save_path = os.path.join(self.get_temp_dir(), filename)
        self.model.save(save_path, save_format=save_format)
        restored_model = keras.models.load_model(save_path)

        # Check we got the real object back.
        self.assertIsInstance(restored_model, AlbertBackbone)

        # Check that output matches.
        restored_output = restored_model(self.input_batch)
        self.assertAllClose(
            model_output["pooled_output"], restored_output["pooled_output"]
        )



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FNet tokenizer."""

import copy
import os

from tensorflow import keras

from keras_nlp.models.f_net.f_net_presets import backbone_presets
from keras_nlp.tokenizers.sentence_piece_tokenizer import SentencePieceTokenizer
from keras_nlp.utils.python_utils import classproperty
from keras_nlp.utils.python_utils import format_docstring


@keras.utils.register_keras_serializable(package="keras_nlp")
class FNetTokenizer(SentencePieceTokenizer):
    """FNet tokenizer layer based on SentencePiece.

    This tokenizer class will tokenize raw strings into integer sequences and
    is based on `keras_nlp.tokenizers.SentencePieceTokenizer`. Unlike the
    underlying tokenizer, it will check for all special tokens needed by
    FNet models and provides a `from_preset()` method to automatically
    download a matching vocabulary for a FNet preset.

    This tokenizer does not provide truncation or padding of inputs. It can be
    combined with a `keras_nlp.models.FNetPreprocessor` layer for input
    packing.

    If input is a batch of strings (rank > 0), the layer will output a
    `tf.RaggedTensor` where the last dimension of the output is ragged.

    If input is a scalar string (rank == 0), the layer will output a dense
    `tf.Tensor` with static shape `[None]`.

    Args:
        proto: Either a `string` path to a SentencePiece proto file, or a
            `bytes` object with a serialized SentencePiece proto. See the
            [SentencePiece repository](https://github.com/google/sentencepiece)
            for more details on the format.

    Examples:

    ```python
    tokenizer = keras_nlp.models.FNetTokenizer(proto="model.spm")

    # Batched inputs.
    tokenizer(["the quick brown fox", "the earth is round"])

    # Unbatched inputs.
    tokenizer("the quick brown fox")

    # Detokenization.
    tokenizer.detokenize(tf.constant([[2, 14, 2231, 886, 2385, 3]]))
    ```
    """

    def __init__(self, proto, **kwargs):
        super().__init__(proto=proto, **kwargs)

        # Check for necessary special tokens.
        cls_token = "[CLS]"
        sep_token = "[SEP]"
        pad_token = "<pad>"
        for token in [cls_token, sep_token, pad_token]:
            if token not in self.get_vocabulary():
                raise ValueError(
                    f"Cannot find token `'{token}'` in the provided "
                    f"`vocabulary`. Please provide `'{token}'` in your "
                    "`vocabulary` or use a pretrained `vocabulary` name."
                )

        self.cls_token_id = self.token_to_id(cls_token)
        self.sep_token_id = self.token_to_id(sep_token)
        self.pad_token_id = self.token_to_id(pad_token)

    @classproperty
    def presets(cls):
        return copy.deepcopy(backbone_presets)

    @classmethod
    @format_docstring(names=", ".join(backbone_presets))
    def from_preset(
        cls,
        preset,
        **kwargs,
    ):
        """Instantiate an FNet tokenizer from preset vocabulary.

        Args:
            preset: string. Must be one of {{names}}.

        Examples:
        ```python
        # Load a preset tokenizer.
        tokenizer = keras_nlp.models.FNetTokenizer.from_preset(
            "f_net_base_en",
        )

        # Tokenize some input.
        tokenizer("The quick brown fox tripped.")

        # Detokenize some input.
        tokenizer.detokenize([5, 6, 7, 8, 9])
        ```
        """
        if preset not in cls.presets:
            raise ValueError(
                "`preset` must be one of "
                f"""{", ".join(cls.presets)}. Received: {preset}."""
            )
        metadata = cls.presets[preset]

        spm_proto = keras.utils.get_file(
            "vocab.spm",
            metadata["spm_proto_url"],
            cache_subdir=os.path.join("models", preset),
            file_hash=metadata["spm_proto_hash"],
        )

        config = metadata["preprocessor_config"]
        config.update(
            {
                "proto": spm_proto,
            },
        )

        return cls.from_config({**config, **kwargs})



# Copyright 2023 The KerasNLP Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GPT2 Causal LM preprocessor layer."""

from absl import logging
from tensorflow import keras

from keras_nlp.models.gpt2.gpt2_preprocessor import GPT2Preprocessor
from keras_nlp.utils.keras_utils import pack_x_y_sample_weight


@keras.utils.register_keras_serializable(package="keras_nlp")
class GPT2CausalLMPreprocessor(GPT2Preprocessor):
    """GPT2 Causal LM preprocessor.

    This preprocessor is majorly used as the preprocesor for `GPT2CausalLM`.
    This class subclasses `keras_nlp.models.GPT2Preprocessor` and keeps most of
    its functionality. The only change is `GPT2CausalLMPreprocessor` sets
    `y` (label) and `sample_weights` field by shifting the input sequence one
    step towards left, and drop the last token as it does not have a successor,
    e.g., if the tokenized input is `[1, 2, 3, 0, 0]` with
    `padding_mask = [1, 1, 1, 0, 0]`, then after preprocessing, we
    will have `x = [1, 2, 3, 0]` and `y = [2, 3, 0, 0]`, with
    `padding_mask = [1, 1, 1, 0]` and `sample_weights = [1, 1, 0, 0]`.

    Args:
        tokenizer: A `keras_nlp.models.GPT2Tokenizer` instance.
        sequence_length: The length of the packed inputs.

    Examples:
    ```python
    # Load the preprocessor from a preset.
    preprocessor = keras_nlp.models.GPT2CausalLMPreprocessor.from_preset(
        "gpt2_base_en"
    )

    # Tokenize and pack a single sentence.
    sentence = tf.constant("League of legends")
    preprocessor(sentence)
    # Same output.
    preprocessor("League of legends")

    # Tokenize a batch of sentences.
    sentences = tf.constant(["Taco tuesday", "Fish taco please!"])
    preprocessor(sentences)
    # Same output.
    preprocessor(["Taco tuesday", "Fish taco please!"])

    # Map a dataset to preprocess a single sentence.
    features = tf.constant(
        [
            "Avatar 2 is amazing!",
            "Well, I am not sure.",
        ]
    )
    labels = tf.constant([1, 0])
    ds = tf.data.Dataset.from_tensor_slices((features, labels))
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)

    # Map a dataset to preprocess unlabled sentences.
    ds = tf.data.Dataset.from_tensor_slices(features)
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)
    """

    def call(self, x, y=None, sample_weight=None):
        if y is not None or sample_weight is not None:
            logging.warning(
                "`GPT2CausalLMPreprocessor` generates `y` and `sample_weight` "
                "based on your input data, but your data already contains `y` "
                "or `sample_weight`. Your `y` and `sample_weight` will be "
                "ignored."
            )

        x = super().call(x)
        token_ids, padding_mask = x["token_ids"], x["padding_mask"]
        # The last token does not have a next token, so we truncate it out.
        x = {
            "token_ids": token_ids[..., :-1],
            "padding_mask": padding_mask[..., :-1],
        }
        # Target `y` will be the next token.
        y = token_ids[..., 1:]
        sample_weight = padding_mask[..., 1:]
        return pack_x_y_sample_weight(x, y, sample_weight)
